@@ -1,7 +1,7 @@
 /*
  *  remove.c
  *
- *  Copyright (c) 2006-2009 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2010 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -51,50 +51,23 @@ int pacman_remove(alpm_list_t *targets)
 	}
 
 	/* Step 0: create a new transaction */
-	if(trans_init(PM_TRANS_TYPE_REMOVE, config->flags) == -1) {
+	if(trans_init(config->flags) == -1) {
 		return(1);
 	}
 
 	/* Step 1: add targets to the created transaction */
 	for(i = targets; i; i = alpm_list_next(i)) {
-		char *targ = alpm_list_getdata(i);
-		if(alpm_trans_addtarget(targ) == -1) {
-			if(pm_errno == PM_ERR_PKG_NOT_FOUND) {
-				printf(_("%s not found, searching for group...\n"), targ);
-				pmgrp_t *grp = alpm_db_readgrp(db_local, targ);
-				if(grp == NULL) {
-					pm_fprintf(stderr, PM_LOG_ERROR, _("'%s': not found in local db\n"), targ);
-					retval = 1;
-					goto cleanup;
-				} else {
-					alpm_list_t *p, *pkgnames = NULL;
-					/* convert packages to package names */
-					for(p = alpm_grp_get_pkgs(grp); p; p = alpm_list_next(p)) {
-						pmpkg_t *pkg = alpm_list_getdata(p);
-						pkgnames = alpm_list_add(pkgnames, (void *)alpm_pkg_get_name(pkg));
-					}
-					printf(_(":: group %s:\n"), targ);
-					list_display("   ", pkgnames);
-					int all = yesno(_("    Remove whole content?"));
-					for(p = pkgnames; p; p = alpm_list_next(p)) {
-						char *pkgn = alpm_list_getdata(p);
-						if(all || yesno(_(":: Remove %s from group %s?"), pkgn, targ)) {
-							if(alpm_trans_addtarget(pkgn) == -1) {
-								pm_fprintf(stderr, PM_LOG_ERROR, "'%s': %s\n", targ,
-								           alpm_strerrorlast());
-								retval = 1;
-								alpm_list_free(pkgnames);
-								goto cleanup;
-							}
-						}
-					}
-					alpm_list_free(pkgnames);
-				}
-			} else {
-				pm_fprintf(stderr, PM_LOG_ERROR, "'%s': %s\n", targ, alpm_strerrorlast());
-				retval = 1;
-				goto cleanup;
-			}
+		char *target = alpm_list_getdata(i);
+		char *targ = strchr(target, '/');
+		if(targ && strncmp(target, "local", 5) == 0) {
+			targ++;
+		} else {
+			targ = target;
+		}
+		if(alpm_remove_target(targ) == -1) {
+			pm_fprintf(stderr, PM_LOG_ERROR, "'%s': %s\n", targ, alpm_strerrorlast());
+			retval = 1;
+			goto cleanup;
 		}
 	}
 
@@ -103,6 +76,12 @@ int pacman_remove(alpm_list_t *targets)
 		pm_fprintf(stderr, PM_LOG_ERROR, _("failed to prepare transaction (%s)\n"),
 		        alpm_strerrorlast());
 		switch(pm_errno) {
+			case PM_ERR_PKG_INVALID_ARCH:
+				for(i = data; i; i = alpm_list_next(i)) {
+					char *pkg = alpm_list_getdata(i);
+					printf(_(":: package %s does not have a valid architecture\n"), pkg);
+				}
+				break;
 			case PM_ERR_UNSATISFIED_DEPS:
 				for(i = data; i; i = alpm_list_next(i)) {
 					pmdepmissing_t *miss = alpm_list_getdata(i);
@@ -123,7 +102,7 @@ int pacman_remove(alpm_list_t *targets)
 
 	/* Search for holdpkg in target list */
 	int holdpkg = 0;
-	for(i = alpm_trans_get_pkgs(); i; i = alpm_list_next(i)) {
+	for(i = alpm_trans_get_remove(); i; i = alpm_list_next(i)) {
 		pmpkg_t *pkg = alpm_list_getdata(i);
 		if(alpm_list_find_str(config->holdpkg, alpm_pkg_get_name(pkg))) {
 			pm_printf(PM_LOG_WARNING, _("%s is designated as a HoldPkg.\n"),
@@ -136,23 +115,26 @@ int pacman_remove(alpm_list_t *targets)
 		goto cleanup;
 	}
 
-	/* Warn user in case of dangerous operation */
-	if(config->flags & PM_TRANS_FLAG_RECURSE ||
-	   config->flags & PM_TRANS_FLAG_CASCADE) {
-		/* list transaction targets */
-		alpm_list_t *pkglist = alpm_trans_get_pkgs();
+	/* Step 3: actually perform the removal */
+	alpm_list_t *pkglist = alpm_trans_get_remove();
+	if(pkglist == NULL) {
+		printf(_(" there is nothing to do\n"));
+		goto cleanup; /* we are done */
+	}
 
-		display_targets(pkglist, 0);
-		printf("\n");
+	if(config->print) {
+		print_packages(pkglist);
+		goto cleanup;
+	}
 
-		/* get confirmation */
-		if(yesno(_("Do you want to remove these packages?")) == 0) {
-			retval = 1;
-			goto cleanup;
-		}
+	/* print targets and ask user confirmation */
+	display_targets(pkglist, 0);
+	printf("\n");
+	if(yesno(_("Do you want to remove these packages?")) == 0) {
+		retval = 1;
+		goto cleanup;
 	}
 
-	/* Step 3: actually perform the removal */
 	if(alpm_trans_commit(NULL) == -1) {
 		pm_fprintf(stderr, PM_LOG_ERROR, _("failed to commit transaction (%s)\n"),
 		        alpm_strerrorlast());
