@@ -1,7 +1,7 @@
 /*
  *  util.c
  *
- *  Copyright (c) 2006-2009 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2010 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -45,10 +45,17 @@
 #include "callback.h"
 
 
-int trans_init(pmtranstype_t type, pmtransflag_t flags)
+int trans_init(pmtransflag_t flags)
 {
-	if(alpm_trans_init(type, flags, cb_trans_evt,
-				cb_trans_conv, cb_trans_progress) == -1) {
+	int ret;
+	if(config->print) {
+		ret = alpm_trans_init(flags, NULL, NULL, NULL);
+	} else {
+		ret = alpm_trans_init(flags, cb_trans_evt, cb_trans_conv,
+				cb_trans_progress);
+	}
+
+	if(ret == -1) {
 		pm_fprintf(stderr, PM_LOG_ERROR, _("failed to init transaction (%s)\n"),
 				alpm_strerrorlast());
 		if(pm_errno == PM_ERR_HANDLE_LOCK) {
@@ -72,83 +79,36 @@ int trans_release(void)
 
 int needs_root(void)
 {
-	if(config->op == PM_OP_UPGRADE || config->op == PM_OP_REMOVE || /* -U, -R */
-	   (config->op == PM_OP_SYNC && (config->op_s_clean || config->op_s_sync || /* -Sc, -Sy */
-	      (!config->group && !config->op_s_info && !config->op_q_list /* all other -S combinations, where */
-	        && !config->op_s_search && !config->op_s_printuris)))) {  /* -g, -i, -l, -s, -p is not set */
-		return(1);
-	} else {
-		return(0);
+	switch(config->op) {
+		case PM_OP_DATABASE:
+			return(1);
+		case PM_OP_UPGRADE:
+		case PM_OP_REMOVE:
+			return(!config->print);
+		case PM_OP_SYNC:
+			return(config->op_s_clean || config->op_s_sync ||
+					(!config->group && !config->op_s_info && !config->op_q_list &&
+					 !config->op_s_search && !config->print));
+		default:
+			return(0);
 	}
 }
 
 /* gets the current screen column width */
 int getcols(void)
 {
-	if(!isatty(1)) {
-		/* We will default to 80 columns if we're not a tty
-		 * this seems a fairly standard file width.
-		 */
-		return 80;
-	} else {
 #ifdef TIOCGSIZE
-		struct ttysize win;
-		if(ioctl(1, TIOCGSIZE, &win) == 0) {
-			return win.ts_cols;
-		}
+	struct ttysize win;
+	if(ioctl(1, TIOCGSIZE, &win) == 0) {
+		return win.ts_cols;
+	}
 #elif defined(TIOCGWINSZ)
-		struct winsize win;
-		if(ioctl(1, TIOCGWINSZ, &win) == 0) {
-			return win.ws_col;
-		}
+	struct winsize win;
+	if(ioctl(1, TIOCGWINSZ, &win) == 0) {
+		return win.ws_col;
+	}
 #endif
-		/* If we can't figure anything out, we'll just assume 80 columns */
-		/* TODO any problems caused by this assumption? */
-		return 80;
-	}
-	/* Original envvar way - prone to display issues
-	const char *cenv = getenv("COLUMNS");
-	if(cenv != NULL) {
-		return atoi(cenv);
-	}
-	return -1;
-	*/
-}
-
-/* does the same thing as 'mkdir -p' */
-int makepath(const char *path)
-{
-	/* A bit of pointer hell here. Descriptions:
-	 * orig - a copy of path so we can safely butcher it with strsep
-	 * str - the current position in the path string (after the delimiter)
-	 * ptr - the original position of str after calling strsep
-	 * incr - incrementally generated path for use in access/mkdir call
-	 */
-	char *orig, *str, *ptr, *incr;
-	mode_t oldmask = umask(0000);
-	int ret = 0;
-
-	orig = strdup(path);
-	incr = calloc(strlen(orig) + 1, sizeof(char));
-	str = orig;
-	while((ptr = strsep(&str, "/"))) {
-		if(strlen(ptr)) {
-			/* we have another path component- append the newest component to
-			 * existing string and create one more level of dir structure */
-			strcat(incr, "/");
-			strcat(incr, ptr);
-			if(access(incr, F_OK)) {
-				if(mkdir(incr, 0755)) {
-					ret = 1;
-					break;
-				}
-			}
-		}
-	}
-	free(orig);
-	free(incr);
-	umask(oldmask);
-	return(ret);
+	return 0;
 }
 
 /* does the same thing as 'rm -rf' */
@@ -257,16 +217,23 @@ void indentprint(const char *str, int indent)
 		return;
 	}
 
+	cols = getcols();
+
+	/* if we're not a tty, print without indenting */
+	if(cols == 0) {
+		printf("%s", str);
+		return;
+	}
+
 	len = strlen(str) + 1;
 	wcstr = calloc(len, sizeof(wchar_t));
 	len = mbstowcs(wcstr, str, len);
 	p = wcstr;
 	cidx = indent;
 
-	if(!p) {
+	if(!p || !len) {
 		return;
 	}
-	cols = getcols();
 
 	while(*p) {
 		if(*p == L' ') {
@@ -285,7 +252,7 @@ void indentprint(const char *str, int indent)
 			}
 			if(len > (cols - cidx - 1)) {
 				/* wrap to a newline and reindent */
-				fprintf(stdout, "\n%-*s", indent, "");
+				printf("\n%-*s", indent, "");
 				cidx = indent;
 			} else {
 				printf(" ");
@@ -293,7 +260,7 @@ void indentprint(const char *str, int indent)
 			}
 			continue;
 		}
-		fprintf(stdout, "%lc", (wint_t)*p);
+		printf("%lc", (wint_t)*p);
 		cidx += wcwidth(*p);
 		p++;
 	}
@@ -307,7 +274,7 @@ char *strtoupper(char *str)
 	char *ptr = str;
 
 	while(*ptr) {
-		(*ptr) = toupper(*ptr);
+		(*ptr) = toupper((unsigned char)*ptr);
 		ptr++;
 	}
 	return str;
@@ -324,7 +291,7 @@ char *strtrim(char *str)
 		return(str);
 	}
 
-	while(isspace(*pch)) {
+	while(isspace((unsigned char)*pch)) {
 		pch++;
 	}
 	if(pch != str) {
@@ -337,7 +304,7 @@ char *strtrim(char *str)
 	}
 
 	pch = (str + (strlen(str) - 1));
-	while(isspace(*pch)) {
+	while(isspace((unsigned char)*pch)) {
 		pch--;
 	}
 	*++pch = '\0';
@@ -345,48 +312,65 @@ char *strtrim(char *str)
 	return(str);
 }
 
-/* Helper function for strreplace */
-static void _strnadd(char **str, const char *append, unsigned int count)
-{
-	if(*str) {
-		*str = realloc(*str, strlen(*str) + count + 1);
-	} else {
-		*str = calloc(sizeof(char), count + 1);
-	}
-
-	strncat(*str, append, count);
-}
-
 /* Replace all occurances of 'needle' with 'replace' in 'str', returning
  * a new string (must be free'd) */
 char *strreplace(const char *str, const char *needle, const char *replace)
 {
-	const char *p, *q;
-	p = q = str;
+	const char *p = NULL, *q = NULL;
+	char *newstr = NULL, *newp = NULL;
+	alpm_list_t *i = NULL, *list = NULL;
+	size_t needlesz = strlen(needle), replacesz = strlen(replace);
+	size_t newsz;
 
-	char *newstr = NULL;
-	unsigned int needlesz = strlen(needle),
-							 replacesz = strlen(replace);
-
-	while (1) {
-		q = strstr(p, needle);
-		if(!q) { /* not found */
-			if(*p) {
-				/* add the rest of 'p' */
-				_strnadd(&newstr, p, strlen(p));
-			}
-			break;
-		} else { /* found match */
-			if(q > p){
-				/* add chars between this occurance and last occurance, if any */
-				_strnadd(&newstr, p, q - p);
-			}
-			_strnadd(&newstr, replace, replacesz);
-			p = q + needlesz;
-		}
+	if(!str) {
+		return(NULL);
 	}
 
-	return newstr;
+	p = str;
+	q = strstr(p, needle);
+	while(q) {
+		list = alpm_list_add(list, (char *)q);
+		p = q + needlesz;
+		q = strstr(p, needle);
+	}
+
+	/* no occurences of needle found */
+	if(!list) {
+		return(strdup(str));
+	}
+	/* size of new string = size of old string + "number of occurences of needle"
+	 * x "size difference between replace and needle" */
+	newsz = strlen(str) + 1 +
+		alpm_list_count(list) * (replacesz - needlesz);
+	newstr = malloc(newsz);
+	if(!newstr) {
+		return(NULL);
+	}
+	*newstr = '\0';
+
+	p = str;
+	newp = newstr;
+	for(i = list; i; i = alpm_list_next(i)) {
+		q = alpm_list_getdata(i);
+		if(q > p){
+			/* add chars between this occurence and last occurence, if any */
+			strncpy(newp, p, q - p);
+			newp += q - p;
+		}
+		strncpy(newp, replace, replacesz);
+		newp += replacesz;
+		p = q + needlesz;
+	}
+	alpm_list_free(list);
+
+	if(*p) {
+		/* add the rest of 'p' */
+		strcpy(newp, p);
+		newp += strlen(p);
+	}
+	*newp = '\0';
+
+	return(newstr);
 }
 
 /** Splits a string into a list of strings using the chosen character as
@@ -446,13 +430,13 @@ void string_display(const char *title, const char *string)
 	int len = 0;
 
 	if(title) {
-		/* compute the length of title + a space */
-		len = string_length(title) + 1;
 		printf("%s ", title);
 	}
 	if(string == NULL || string[0] == '\0') {
 		printf(_("None"));
 	} else {
+		/* compute the length of title + a space */
+		len = string_length(title) + 1;
 		indentprint(string, len);
 	}
 	printf("\n");
@@ -474,18 +458,20 @@ void list_display(const char *title, const alpm_list_t *list)
 		for(i = list, cols = len; i; i = alpm_list_next(i)) {
 			char *str = alpm_list_getdata(i);
 			int s = string_length(str);
-			/* two additional spaces are added to the length */
-			s += 2;
 			int maxcols = getcols();
-			if(s + cols > maxcols) {
+			if(maxcols > 0 && (cols + s + 2) >= maxcols) {
 				int j;
 				cols = len;
 				printf("\n");
 				for (j = 1; j <= len; j++) {
 					printf(" ");
 				}
+			} else if (cols != len) {
+				/* 2 spaces are added if this is not the first element on a line. */
+				printf("  ");
+				cols += 2;
 			}
-			printf("%s  ", str);
+			printf("%s", str);
 			cols += s;
 		}
 		printf("\n");
@@ -564,7 +550,9 @@ void display_targets(const alpm_list_t *pkgs, int install)
 		printf("\n");
 
 		printf(_("Total Download Size:    %.2f MB\n"), mbdlsize);
-		printf(_("Total Installed Size:   %.2f MB\n"), mbisize);
+		if(!(config->flags & PM_TRANS_FLAG_DOWNLOADONLY)) {
+			printf(_("Total Installed Size:   %.2f MB\n"), mbisize);
+		}
 	} else {
 		asprintf(&str, _("Remove (%d):"), alpm_list_count(targets));
 		list_display(str, targets);
@@ -577,35 +565,96 @@ void display_targets(const alpm_list_t *pkgs, int install)
 	FREELIST(targets);
 }
 
-/* Display a list of transaction targets.
- * `pkgs` should be a list of pmpkg_t's,
- * retrieved from a transaction object
- */
-void display_synctargets(const alpm_list_t *syncpkgs)
+static off_t pkg_get_size(pmpkg_t *pkg)
 {
-	const alpm_list_t *i, *j;
-	alpm_list_t *pkglist = NULL, *rpkglist = NULL;
-
-	for(i = syncpkgs; i; i = alpm_list_next(i)) {
-		pmpkg_t *pkg = alpm_list_getdata(i);
-		pkglist = alpm_list_add(pkglist, pkg);
-
-		/* The removes member contains a list of packages to be removed
-		 * due to the package that is being installed. */
-		alpm_list_t *to_replace = alpm_pkg_get_removes(pkg);
-
-		for(j = to_replace; j; j = alpm_list_next(j)) {
-			pmpkg_t *rp = alpm_list_getdata(j);
-			rpkglist = alpm_list_add(rpkglist, rp);
-		}
+	switch(config->op) {
+		case PM_OP_SYNC:
+			return(alpm_pkg_download_size(pkg));
+		case PM_OP_UPGRADE:
+			return(alpm_pkg_get_size(pkg));
+		default:
+			return(alpm_pkg_get_isize(pkg));
 	}
+}
 
-	/* start displaying information */
-	display_targets(rpkglist, 0);
-	display_targets(pkglist, 1);
+static char *pkg_get_location(pmpkg_t *pkg)
+{
+	pmdb_t *db;
+	const char *dburl;
+	char *string;
+	switch(config->op) {
+		case PM_OP_SYNC:
+			db = alpm_pkg_get_db(pkg);
+			dburl = alpm_db_get_url(db);
+			if(dburl) {
+				char *pkgurl = NULL;
+				asprintf(&pkgurl, "%s/%s", dburl, alpm_pkg_get_filename(pkg));
+				return(pkgurl);
+			}
+		case PM_OP_UPGRADE:
+			return(strdup(alpm_pkg_get_filename(pkg)));
+		default:
+			string = NULL;
+			asprintf(&string, "%s-%s", alpm_pkg_get_name(pkg), alpm_pkg_get_version(pkg));
+			return(string);
+	}
+}
 
-	alpm_list_free(pkglist);
-	alpm_list_free(rpkglist);
+void print_packages(const alpm_list_t *packages)
+{
+	const alpm_list_t *i;
+	if(!config->print_format) {
+		config->print_format = strdup("%l");
+	}
+	for(i = packages; i; i = alpm_list_next(i)) {
+		pmpkg_t *pkg = alpm_list_getdata(i);
+		char *string = strdup(config->print_format);
+		char *temp = string;
+		/* %n : pkgname */
+		if(strstr(temp,"%n")) {
+			string = strreplace(temp, "%n", alpm_pkg_get_name(pkg));
+			free(temp);
+			temp = string;
+		}
+		/* %v : pkgver */
+		if(strstr(temp,"%v")) {
+			string = strreplace(temp, "%v", alpm_pkg_get_version(pkg));
+			free(temp);
+			temp = string;
+		}
+		/* %l : location */
+		if(strstr(temp,"%l")) {
+			char *pkgloc = pkg_get_location(pkg);
+			string = strreplace(temp, "%l", pkgloc);
+			free(pkgloc);
+			free(temp);
+			temp = string;
+		}
+		/* %r : repo */
+		if(strstr(temp,"%r")) {
+			const char *repo = "local";
+			pmdb_t *db = alpm_pkg_get_db(pkg);
+			if(db) {
+				repo = alpm_db_get_name(db);
+			}
+			string = strreplace(temp, "%r", repo);
+			free(temp);
+			temp = string;
+		}
+		/* %s : size */
+		if(strstr(temp,"%s")) {
+			char *size;
+			double mbsize = 0.0;
+			mbsize = pkg_get_size(pkg) / (1024.0 * 1024.0);
+			asprintf(&size, "%.2f", mbsize);
+			string = strreplace(temp, "%s", size);
+			free(size);
+			free(temp);
+			temp = string;
+		}
+		printf("%s\n",string);
+		free(string);
+	}
 }
 
 /* Helper function for comparing strings using the
@@ -640,6 +689,7 @@ void display_optdepends(pmpkg_t *pkg)
 static int question(short preset, char *fmt, va_list args)
 {
 	char response[32];
+	int sresponse = sizeof(response)-1;
 	FILE *stream;
 
 	if(config->noconfirm) {
@@ -662,7 +712,7 @@ static int question(short preset, char *fmt, va_list args)
 		return(preset);
 	}
 
-	if(fgets(response, 32, stdin)) {
+	if(fgets(response, sresponse, stdin)) {
 		strtrim(response);
 		if(strlen(response) == 0) {
 			return(preset);
